#include "pass.h"

/* handle activity on one end of the proxy */
void pass_proxy_cb(mio m, int state, void *arg, char *buffer, int bufsz) {
	proxy pr = (proxy) arg;

	log_debug(ZONE, "activity on proxy, mio %p", m);

	switch(state) {
		case MIO_BUFFER:
			log_debug(ZONE, "MIO_BUFFER, resending data");

			/* data appeared, so forward it on */
			if(m == pr->local)
				mio_write(pr->remote, NULL, buffer, bufsz);
			else if(m == pr->remote)
				mio_write(pr->local, NULL, buffer, bufsz);

			break;

		case MIO_CLOSED:
			log_debug(ZONE, "MIO_CLOSED, closing other end");
			
			/* one end got closed, so close the other end */
			if(m == pr->local) {
				mio_close(pr->remote);
				pr->remote = NULL;
			}
			else if(m == pr->remote) {
				mio_close(pr->local);
				pr->local = NULL;
			}
			break;
	}
}

/* accept and handle a local (owner) connection */
void pass_local_accept_cb(mio lm, int state, void *arg, char *buffer, int bufsz) {
	proxy pr = (proxy) arg;

	/* only interested in new connections */
	if(state != MIO_NEW)
		return;

	/* we don't want anyone else connecting */
	mio_close(pr->local);
	pr->local = NULL;
	
	log_debug(ZONE, "accept on owner mio, port %d (state %d)", pr->local_port, state);

	/* got the other end, give it a handler too */
	mio_reset(lm, pass_proxy_cb, (void *) pr);
	
	pr->local = lm;
	/* FIXME: pr->local_port = */
}

/* accept and handle a remote connection */
void pass_remote_accept_cb(mio rm, int state, void *arg, char *buffer, int bufsz) {
	proxy pr = (proxy) arg;
	passi pi = pr->pi;
	int port, port_first;
	mio lm = NULL;
	jpacket jp;
	char port_s[12];

	/* only interested in new connections */
	if(state != MIO_NEW)
		return;
	
	/* we don't want anyone else connecting */
	mio_close(pr->remote);
	pr->remote = NULL;

	log_debug(ZONE, "accept on pas, port %d (state %d)", pr->remote_port, state);

	/* we have one end of the proxy, so it needs a proper handler */
	mio_reset(rm, pass_proxy_cb, (void *) pr);

	pr->remote = rm;
	/* FIXME: pr->remote_port = */

	/* find a port to listen for our owner connecting */
	port_first = pi->port_next;

	while(lm == NULL) {
		port = pi->port_next;
		lm = mio_listen(port, pi->local_ip, pass_local_accept_cb, (void *) pr, NULL, NULL);

		pi->port_next = (pi->port_next == pi->port_high) ? pi->port_low : pi->port_next + 1;

		if(lm == NULL && pi->port_next == port_first) {
			log_error(ZONE, "couldn't create mio, no ports available");

			return;
		}
	}

	pr->local = lm;
	pr->local_port = port;

	/* let the owner know someone has connected */
	
	jp = jpacket_new(jutil_iqnew(JPACKET__SET, NS_PASS));
	
	xmlnode_put_attrib(jp->x, "to", jid_full(pr->owner));
	xmlnode_put_attrib(jp->x, "from", pr->pi->i->id);

	sprintf(port_s, "%d", pr->local_port);

	xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq, "client"), pr->local->ip, -1);
	/* FIXME: xmlnode_put_attrib(xmlnode_get_tag(jp->iq, "client"), "port", port_s); */

	sprintf(port_s, "%d", pr->remote_port);

	xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq, "proxy"), pi->remote_ip, -1);
	xmlnode_put_attrib(xmlnode_get_tag(jp->iq, "proxy"), "port", port_s);

	deliver(dpacket_new(jp->x), NULL);
	
	log_debug(ZONE, "new pas connection, waiting for owner on port %d", port);
}

/* handle an iq set - create the first listen port */
void pass_handle_set(passi pi, jpacket jp) {
	mio m = NULL;
	proxy pr;
	int port, port_first;
	char port_s[12];
	
	log_debug(ZONE, "in pass_handle_set, creating new proxy");
	
	/* a new proxy struct */
	pr = (proxy) pmalloco(pi->p, sizeof(_proxy));
	
	pr->pi = pi;
	pr->p = pi->p;
		
	/* find the first spare port within the port range, and use listen on it */
	port_first = pi->port_next;

	while(m == NULL) {
		port = pi->port_next;
		m = mio_listen(port, pi->remote_ip, pass_remote_accept_cb, (void *) pr, NULL, NULL);

		pi->port_next = (pi->port_next == pi->port_high) ? pi->port_low : pi->port_next + 1;

		if(m == NULL && pi->port_next == port_first) {
			log_error(ZONE, "couldn't create mio, no ports available");

			return;
		}
	}

	pr->remote = m;
	pr->remote_port = port;

	pr->owner = jid_new(pr->p, xmlnode_get_attrib(jp->x, "from"));

	/* inform the owner */

	jutil_iqresult(jp->x);
	xmlnode_put_attrib(xmlnode_insert_tag(jp->x, "query"), "xmlns", NS_PASS);
	jpacket_reset(jp);

	sprintf(port_s, "%d", pr->remote_port);

	xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq, "server"), pi->remote_ip, -1);
	xmlnode_put_attrib(xmlnode_get_tag(jp->iq, "server"), "port", port_s);

	deliver(dpacket_new(jp->x), NULL);

	log_debug(ZONE, "new proxy created on ip %s port %d", pi->remote_ip, pr->remote_port);
}

/* actually handle the packet (inside a new thread) */
void _pass_packets(void *arg) {
	jpacket jp = (jpacket) arg;
	passi pi = (passi) jp->aux1;
	xmlnode x = jp->x;
	char *type;

	log_debug(ZONE, "handling packet: %s", xmlnode2str(x));

	/* we're only interested in iq packets */
	if(jp->type == JPACKET_IQ) {

		/* and then, only PASS queries */
		if(j_strcmp(xmlnode_get_attrib(xmlnode_get_tag(x, "query"), "xmlns"), NS_PASS) != 0) {
			jutil_error(jp->x, TERROR_BAD);
			deliver(dpacket_new(jp->x), NULL);

			return;
		}
		
		type = xmlnode_get_attrib(x, "type");

		/* only handle set at this time */
		if(j_strcmp(type, "set") == 0)
			pass_handle_set(pi, jp);

		else {
			jutil_error(jp->x, TERROR_BAD);
			deliver(dpacket_new(jp->x), NULL);

			return;
		}
	}
}

result pass_packets(instance i, dpacket dp, void *arg) {
	passi pi = (passi) arg;
	jpacket jp;

	/* get the packet - it may be a route packet */
	if(dp->type == p_ROUTE)
		jp = jpacket_new(xmlnode_get_firstchild(dp->x));
	else
		jp = jpacket_new(dp->x);

	/* delivery failed */
	if(jp == NULL) {
		deliver_fail(dp, "illegal packet");
		return r_DONE;
	}

	/* bad packet? */
	if(jp->type == JPACKET_UNKNOWN || jp->to == NULL) {
		jutil_error(jp->x, TERROR_BAD);
		deliver(dpacket_new(jp->x), NULL);

		return r_DONE;
	}

	log_debug(ZONE, "queueing packet: %s", xmlnode2str(jp->x));

	jp->aux1 = (void *) pi;

	/* make a new thread for the packet, so we don't block the main
	   thread for too long */
	mtq_send(pi->q, jp->p, _pass_packets, (void *) jp);

	return r_DONE;
};

void pass(instance i, xmlnode x) {
	passi pi;

	log_debug(ZONE, "pass loading");

	/* setup a new pass instance */
	pi = (passi) pmalloco(i->p, sizeof(_passi));
	pi->i = i;
	pi->p = i->p;

	/* new thread queue, so we don't stall too long when processing packets */
	pi->q = mtq_new(i->p);

	/* get our config */
	pi->config = xdb_get(xdb_cache(i), jid_new(xmlnode_pool(x), "config@-internal"), "jabber:config:pass");

	/* ip address */
	pi->local_ip = xmlnode_get_tag_data(pi->config, "local-ip");
	if(pi->local_ip == NULL)
		pi->local_ip = "127.0.0.1";

	pi->remote_ip = xmlnode_get_tag_data(pi->config, "remote-ip");
	if(pi->remote_ip == NULL)
		pi->remote_ip = "127.0.0.1";

	/* port range */
	pi->port_low = j_atoi(xmlnode_get_tag_data(pi->config, "port-low"), 40000);
	pi->port_high = j_atoi(xmlnode_get_tag_data(pi->config, "port-high"), 49999);
	pi->port_next = pi->port_low;

	register_phandler(i, o_DELIVER, pass_packets, (void *) pi);
}
